/*
 *  $Id: vscan-fsav.c,v 1.1.2.9 2005/02/23 02:49:37 reniar Exp $
 *
 * Virusscanning VFS module for samba.  
 * Log infected files via syslog
 * facility and block access using fsecure antivirus
 *
 * Copyright (C) Monex AG  Oliver Jehle, 2003
 * Copyright (C) Rainer Link, 2004
 *               OpenAntiVirus.org <rainer@openantivirus.org>
 *
 * based on the existing  F-PROT module by
 * Copyright (C) Rainer Link, 2001-2003
 * Copyright (C) Stefan (metze) Metzmacher, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "vscan-global.h"
#include "vscan-vfs.h"
#include "vscan-fsav.h"


static const char module_id[] =
    VSCAN_MODULE_STR " " SAMBA_VSCAN_VERSION_STR;

vscan_config_struct vscan_config; /* contains the vscan module configuration */

BOOL fsav_userinstance;		/* start own fsavd for each user */
pstring fsav_config_file;	/* location of config file */
pstring fsav_socket;		/* socketname */
pstring fsav_db_dir;		/* location of the database dir */
pstring fsav_binary;		/* location of the fsavd deamon */
pstring fsav_user;		/* needed only for userinstance socket name */
BOOL fsav_archive;		/* scan archives */
BOOL fsav_mime;			/* scan archives */
uid_t fsav_connect_uid;		/* connect to socket as uid */
int fsav_maxnested;		/* maximum level archives */
int fsav_timeout;		/* scan timeout */

BOOL verbose_file_logging;	/* log ever file access */
BOOL send_warning_message;	/* send a warning message using the windows
				   messenger service? */

static BOOL do_parameter(const char *param, const char *value)
{
        if ( do_common_parameter(&vscan_config, param, value) == False ) {
                /* parse VFS module specific configuration values */
		if (StrCaseCmp("fsav userinstance", param) == 0) {
			DEBUG(3, ("fsav user instance value : %s\n", value));
			set_boolean(&fsav_userinstance, value);
			DEBUG(3,
			      ("fsav user instance  is: %d\n", fsav_userinstance));
		} else if (StrCaseCmp("fsav config file", param) == 0) {
			pstrcpy(fsav_config_file, value);
			DEBUG(3, ("fsav config file is: %s\n", fsav_config_file));
		} else if (StrCaseCmp("fsav binary", param) == 0) {
			pstrcpy(fsav_binary, value);
			DEBUG(3, ("fsav binary: %s\n", fsav_binary));
		} else if (StrCaseCmp("fsav db dir", param) == 0) {
			pstrcpy(fsav_db_dir, value);
			DEBUG(3, ("fsav db dir is: %s\n", fsav_db_dir));
		} else if (StrCaseCmp("fsav socket", param) == 0) {
			pstrcpy(fsav_socket, value);
			DEBUG(3, ("fsav socket is: %s\n", fsav_socket));
		} else if (StrCaseCmp("fsav mime scan", param) == 0) {
			set_boolean(&fsav_mime, value);
			DEBUG(3, ("fsav mime scan is : %d\n", fsav_mime));
		} else if (StrCaseCmp("fsav connect uid", param) == 0) {
			fsav_connect_uid = atoi(value);
			DEBUG(3, ("fsav connect uid is : %i\n", fsav_connect_uid));
		} else if (StrCaseCmp("fsav archive scan", param) == 0) {
			set_boolean(&fsav_archive, value);
			DEBUG(3, ("fsav archive scan is : %d\n", fsav_mime));
		} else if (StrCaseCmp("fsav max nested level", param) == 0) {
			fsav_maxnested = atoi(value);
			DEBUG(3,
			      ("fsav max nested level is : %i\n", fsav_maxnested));
		} else if (StrCaseCmp("fsav timeout", param) == 0) {
			fsav_timeout = atoi(value);
			DEBUG(3, ("fsav timeout is : %i\n", fsav_timeout));
		} else
			DEBUG(3, ("unknown parameter: %s\n", param));
	}

	return True;
}

static BOOL do_section(const char *section)
{
	/* simply return true, there's only one section :-) */
	return True;
}


/* Implementation of vfs_ops.  */

#if (SMB_VFS_INTERFACE_VERSION >= 6)
static int vscan_connect(vfs_handle_struct * handle,
			 connection_struct * conn, const char *svc,
			 const char *user)
#else
static int vscan_connect(struct connection_struct *conn, 
			PROTOTYPE_CONST char *svc, 
			PROTOTYPE_CONST char *user)
#endif
{
        fstring config_file;            /* location of config file, either
                                           PARAMCONF or as set via vfs options
                                        */

        #if (SAMBA_VERSION_MAJOR==2 && SAMBA_VERSION_RELEASE>=4) || SAMBA_VERSION_MAJOR==3
         #if !(SMB_VFS_INTERFACE_VERSION >= 6)
          pstring opts_str;
          PROTOTYPE_CONST char *p;
         #endif
        #endif
        int retval;

#if (SMB_VFS_INTERFACE_VERSION >= 6)
	vscan_syslog
	    ("samba-vscan (%s) connected (Samba 3.0), (c) by Rainer Link, OpenAntiVirus.org",
	     module_id);
#endif

	/* set default value for configuration files */
	fstrcpy(config_file, PARAMCONF);

        /* set default values */
        set_common_default_settings(&vscan_config);
        DEBUG(0, ("max size: %d\n", vscan_config.common.max_size));

	/* fsav specific defaults */

	fsav_userinstance = VSAN_FSAV_USERINSTANCE;
	fsav_connect_uid = VSCAN_FSAV_CONNECT_ID;
	fsav_timeout = VSCAN_FSAV_TIMEOUT;
	fsav_maxnested = VSCAN_FSAV_MAX_NESTED_LEVEL;
	fsav_archive = VSCAN_FSAV_ARCHIVE_SCAN; 
	fsav_mime = VSCAN_FSAV_MINE_SCAN;
	pstrcpy(fsav_config_file, VSCAN_FSAV_CONFIG_FILE);
	pstrcpy(fsav_socket, VSCAN_FSAV_SOCKET);
	pstrcpy(fsav_db_dir, VSCAN_FSAV_DB_DIR);
	pstrcpy(fsav_binary, VSCAN_FSAV_DB_DIR);
	pstrcpy(fsav_user, user);

	vscan_syslog("INFO: connect to service %s by user %s", svc, user);

        #if (SAMBA_VERSION_MAJOR==2 && SAMBA_VERSION_RELEASE>=4) || SAMBA_VERSION_MAJOR==3

         fstrcpy(config_file, get_configuration_file(conn, VSCAN_MODULE_STR, PARAMCONF));
         DEBUG(3, ("configuration file is: %s\n", config_file));

 	 retval = pm_process(config_file, do_section, do_parameter);
	 DEBUG(10, ("pm_process returned %d\n", retval));

         /* FIXME: this is lame! */
         verbose_file_logging = vscan_config.common.verbose_file_logging;
         send_warning_message = vscan_config.common.send_warning_message;

	 if (!retval)
		vscan_syslog
		    ("ERROR: could not parse configuration file '%s'. File not found or not read-able. Using compiled-in defaults",
		     config_file);
	#endif

	/* initialise lrufiles list */
	DEBUG(5, ("init lrufiles list\n"));
	lrufiles_init(vscan_config.common.max_lrufiles, vscan_config.common.lrufiles_invalidate_time);

	/* initialise filetype */
	DEBUG(5, ("init file type\n"));
	filetype_init(0, vscan_config.common.exclude_file_types);

        #if (SMB_VFS_INTERFACE_VERSION >= 6)
         return SMB_VFS_NEXT_CONNECT(handle, conn, svc, user);
        #else
         return default_vfs_ops.connect(conn, svc, user);
        #endif
}

#if (SMB_VFS_INTERFACE_VERSION >= 6)
static void vscan_disconnect(vfs_handle_struct * handle,
			     connection_struct * conn)
#else
static void vscan_disconnect(struct connection_struct *conn)
#endif
{

	vscan_syslog("INFO: disconnected");

	lrufiles_destroy_all();
	filetype_close();

#if (SMB_VFS_INTERFACE_VERSION >= 6)
        SMB_VFS_NEXT_DISCONNECT(handle, conn);
#else
        default_vfs_ops.disconnect(conn);
#endif
}

#if (SMB_VFS_INTERFACE_VERSION >= 6)
static int vscan_open(vfs_handle_struct * handle, connection_struct * conn,
		      const char *fname, int flags, mode_t mode)
#else
static int vscan_open(struct connection_struct *conn, 
			PROTOTYPE_CONST char *fname, int flags, mode_t mode)

#endif
{
	int retval, must_be_checked;
	SMB_STRUCT_STAT stat_buf;
	pstring filepath;
	char client_ip[CLIENT_IP_SIZE];
	fsav_handle *fsav;

	int rc;


	/* Assemble complete file path */
	pstrcpy(filepath, conn->connectpath);
	pstrcat(filepath, "/");
	pstrcat(filepath, fname);

	/* scan files while opening? */
        if ( !vscan_config.common.scan_on_open ) {
                DEBUG(3, ("samba-vscan - open: File '%s' not scanned as scan_on_open is not set\n", fname));
#if (SMB_VFS_INTERFACE_VERSION >= 6)
                return SMB_VFS_NEXT_OPEN(handle, conn, fname, flags, mode);
#else
                return default_vfs_ops.open(conn, fname, flags, mode);
#endif
        }

#if (SMB_VFS_INTERFACE_VERSION >= 6)
        if ( (SMB_VFS_NEXT_STAT(handle, conn, fname, &stat_buf)) != 0 )    /* an error occured */
                return SMB_VFS_NEXT_OPEN(handle, conn, fname, flags, mode);
#else
        if ( (default_vfs_ops.stat(conn, fname, &stat_buf)) != 0 )    /* an error occured */
                return default_vfs_ops.open(conn, fname, flags, mode);
#endif
        else if ( S_ISDIR(stat_buf.st_mode) )   /* is it a directory? */
#if (SMB_VFS_INTERFACE_VERSION >= 6)
                return SMB_VFS_NEXT_OPEN(handle, conn, fname, flags, mode);
#else
                return default_vfs_ops.open(conn, fname, flags, mode);
#endif
        else if ( ( stat_buf.st_size > vscan_config.common.max_size ) && ( vscan_config.common.max_size > 0 ) ) /* file is too large */
                vscan_syslog("INFO: File %s is larger than specified maximum file size! Not scanned!", fname);
        else if ( stat_buf.st_size == 0 ) /* do not scan empty files */
#if (SMB_VFS_INTERFACE_VERSION >= 6)
                return SMB_VFS_NEXT_OPEN(handle, conn, fname, flags, mode);
#else
                return default_vfs_ops.open(conn, fname, flags, mode);
#endif
        else if ( filetype_skipscan(filepath) == VSCAN_FT_SKIP_SCAN ) {
                if ( verbose_file_logging )
                        vscan_syslog("File '%s' not scanned as file type is on exclude list", filepath);
#if (SMB_VFS_INTERFACE_VERSION >= 6)
                return SMB_VFS_NEXT_OPEN(handle, conn, fname, flags, mode);
#else
                return default_vfs_ops.open(conn, fname, flags, mode);
#endif
	} else {
		fsav = fsav_create_handle();
		if (fsav) {
			fsav->userinstance = fsav_userinstance;
			fsav->connect_uid = fsav_connect_uid;
			fsav->maxnested = fsav_maxnested;
			fsav->archive = fsav_archive;
			fsav->timeout = fsav_timeout;
			fsav->mime = fsav_mime;
			fsav->binary = fsav_binary;
			fsav->config_file = fsav_config_file;
			fsav->db_dir = fsav_db_dir;
			fsav->socket = fsav_socket;
			fsav->user = fsav_user;
			if (fsav_connect_handle(fsav) != 0) {
				fsav_free_handle(fsav);
				fsav = NULL;
			}
		}
		if (!fsav && vscan_config.common.deny_access_on_error) {
			/* an error occured - can not communicate to daemon - deny access */
			vscan_syslog
			    ("ERROR: can not communicate to daemon - access denied");
			errno = EACCES;
			return -1;
		} else {

			/* get client IP */
			safe_strcpy(client_ip, conn->client_address,
				    CLIENT_IP_SIZE - 1);

			/* must file actually be scanned? */
			must_be_checked =
			    lrufiles_must_be_checked(filepath,
						     stat_buf.st_mtime);
			if (must_be_checked == VSCAN_LRU_DENY_ACCESS) {
				/* file has already been checked and marked as infected */
				/* deny access */
				if ( vscan_config.common.verbose_file_logging )
					vscan_syslog
					    ("File '%s' has already been scanned and marked as infected. Not scanned any more. Access denied",
					     filepath);

				/* close handle */
				fsav_free_handle(fsav);

				/* deny access */
				errno = EACCES;
				return -1;
			} else if (must_be_checked == VSCAN_LRU_GRANT_ACCESS) {
				/* file has already been checked, 
				   not marked as infected and not modified */
				if ( vscan_config.common.verbose_file_logging )
					vscan_syslog
					    ("File '%s' has already been scanned, not marked as infected and not modified. Not scanned anymore. Access granted",
					     filepath);

				/* close handle */
				fsav_free_handle(fsav);

				/* grant access */
#if (SMB_VFS_INTERFACE_VERSION >= 6)
				return SMB_VFS_NEXT_OPEN(handle, conn,
							 fname, flags,
							 mode);
#else
				return default_vfs_ops.open(conn, fname, flags, mode);
#endif
			}
			/* ok, we must check the file */

			/* scan file */
			retval = fsav_scan(fsav, filepath, client_ip);
			if (retval == VSCAN_SCAN_MINOR_ERROR && vscan_config.common.deny_access_on_minor_error) {
				/* a minor error occured - deny access */
				vscan_syslog
				    ("ERROR: daemon failed with a minor error - access to file %s denied",
				     fname);

				/* close handle */
				fsav_free_handle(fsav);

				/* to be safe, remove file from lrufiles */
				lrufiles_delete(filepath);

				/* deny access */
				errno = EACCES;
				return -1;
			} else if (retval == VSCAN_SCAN_ERROR && vscan_config.common.deny_access_on_error) {
				/* an error occured - can not communicate to daemon - deny access */
				vscan_syslog
				    ("ERROR: can not communicate to daemon - access to file %s denied",
				     fname);

				/* close handle */
				fsav_free_handle(fsav);

				/* to be safe, remove file from lrufiles */
				lrufiles_delete(filepath);

				/* deny access */

				errno = EACCES;
				return -1;
			} else if (retval == VSCAN_SCAN_VIRUS_FOUND) {
				fsav_free_handle(fsav);
				/* do action ... */
#if (SMB_VFS_INTERFACE_VERSION >= 6)
				rc = vscan_do_infected_file_action(handle,
								   conn,
								   filepath,
								   vscan_config.common.quarantine_dir,
								   vscan_config.common.quarantine_prefix,
						   		   vscan_config.common.infected_file_action);
#else
				rc = rc = vscan_do_infected_file_action(&default_vfs_ops,
                                                                   conn,
                                                                   filepath,
                                                                   vscan_config.common.quarantine_dir,
                                                                   vscan_config.common.quarantine_prefix,
                                                                   vscan_config.common.infected_file_action);
#endif

				/* add/update file. mark file as infected! */
				lrufiles_add(filepath, stat_buf.st_mtime,
					     True);

				/* virus found, deny acces */
				errno = EACCES;
				return -1;
			} else if (retval == VSCAN_SCAN_OK) {
				/* file is clean, add to lrufiles */
				lrufiles_add(filepath, stat_buf.st_mtime,
					     False);
			}
		}

		/* close handle */
		fsav_free_handle(fsav);
	}
#if (SMB_VFS_INTERFACE_VERSION >= 6)
       return SMB_VFS_NEXT_OPEN(handle, conn, fname, flags, mode);
#else
       return default_vfs_ops.open(conn, fname, flags, mode);
#endif

}

#if (SMB_VFS_INTERFACE_VERSION >= 6)
static int vscan_close(vfs_handle_struct * handle, files_struct * fsp,
		       int fd)
#else
static int vscan_close(struct files_struct *fsp, int fd)
#endif
{
	fsav_handle *fsav;
	pstring filepath;
	int retval, rv, rc;
	char client_ip[CLIENT_IP_SIZE];

	/* First close the file */
#if (SMB_VFS_INTERFACE_VERSION >= 6)
        retval = SMB_VFS_NEXT_CLOSE(handle, fsp, fd);
#else
        retval = default_vfs_ops.close(fsp, fd);
#endif

	if (!vscan_config.common.scan_on_close) {
		DEBUG(3,
		      ("samba-vscan - close: File '%s' not scanned as scan_on_close is not set\n",
		       fsp->fsp_name));
		return retval;
	}


	/* get the file name */
	pstrcpy(filepath, fsp->conn->connectpath);
	pstrcat(filepath, "/");
	pstrcat(filepath, fsp->fsp_name);

	/* Don't scan directorys */
	if (fsp->is_directory)
		return retval;


	if (!fsp->modified) {
		if ( vscan_config.common.verbose_file_logging )
			vscan_syslog
			    ("INFO: file %s was not modified - not scanned",
			     filepath);

		return retval;
	}

	/* don't scan files which are in the list of exclude file types */
	if (filetype_skipscan(filepath) == VSCAN_FT_SKIP_SCAN) {
		if ( vscan_config.common.verbose_file_logging )
			vscan_syslog
			    ("File '%s' not scanned as file type is on exclude list",
			     filepath);
		return retval;
	}


	/* scan only file, do nothing */

	fsav = fsav_create_handle();
	if (fsav) {
		fsav->userinstance = fsav_userinstance;
		fsav->connect_uid = fsav_connect_uid;
		fsav->maxnested = fsav_maxnested;
		fsav->archive = fsav_archive;
		fsav->timeout = fsav_timeout;
		fsav->mime = fsav_mime;
		fsav->binary = fsav_binary;
		fsav->config_file = fsav_config_file;
		fsav->db_dir = fsav_db_dir;
		fsav->socket = fsav_socket;
		fsav->user = fsav_user;
		if (fsav_connect_handle(fsav) != 0) {
			fsav_free_handle(fsav);
			fsav = NULL;
		}

	}
	if (fsav) {
		safe_strcpy(client_ip, fsp->conn->client_address,
			    CLIENT_IP_SIZE - 1);
		rv = fsav_scan(fsav, filepath, client_ip);
		if (rv == VSCAN_SCAN_VIRUS_FOUND) {
			/* virus was found */
#if (SMB_VFS_INTERFACE_VERSION >= 6)
			rc = vscan_do_infected_file_action(handle,
							   fsp->conn,
							   filepath,
							   vscan_config.common.quarantine_dir,
							   vscan_config.common.quarantine_prefix,
							   vscan_config.common.infected_file_action);
#else
			rc = vscan_do_infected_file_action(&default_vfs_ops,
                                                           fsp->conn,
                                                           filepath,
                                                           vscan_config.common.quarantine_dir,
                                                           vscan_config.common.quarantine_prefix,
                                                           vscan_config.common.infected_file_action);
#endif
		}
		/* FIXME: shouldn't we handle errors as well? */

	}
	return retval;
}

#if (SMB_VFS_INTERFACE_VERSION >= 6)
/* Samba 3.0 */
NTSTATUS init_module(void)
{
	NTSTATUS ret;

	ret =
	    smb_register_vfs(SMB_VFS_INTERFACE_VERSION, VSCAN_MODULE_STR,
			     vscan_ops);
	DEBUG(5,
	      ("samba-vscan (%s) registered (Samba 3.0), (c) by Rainer Link, OpenAntiVirus.org\n",
	       module_id));
	openlog("smbd_" VSCAN_MODULE_STR, LOG_PID, SYSLOG_FACILITY);

	return ret;
}
#else
/* VFS initialisation function.  Return initialised vfs_ops structure
   back to SAMBA. */
#if SAMBA_VERSION_MAJOR==3
 /* Samba 3.0 alphaX */
 vfs_op_tuple *vfs_init(int *vfs_version, struct vfs_ops *def_vfs_ops,
                        struct smb_vfs_handle_struct *vfs_handle)
#else
 /* Samba 2.2.x */
 #if SAMBA_VERSION_RELEASE>=4
  /* Samba 2.2.4 */
  struct vfs_ops *vfs_init(int *vfs_version, struct vfs_ops *def_vfs_ops)
 #elif SAMBA_VERSION_RELEASE==2
  /* Samba 2.2.2 / Samba 2.2.3 !!! */
  struct vfs_ops *vfs_init(int* Version, struct vfs_ops *ops)
 #elif SAMBA_VERSION_RELEASE==1
  /* Samba 2.2.1 */
  struct vfs_ops *vfs_module_init(int *vfs_version)
 #else
  /* Samba 2.2.0 */
  struct vfs_ops *vfs_init(int *vfs_version)
 #endif
#endif
{
        #if SAMBA_VERSION_MAJOR!=3
         #if SAMBA_VERSION_RELEASE>=4
          /* Samba 2.2.4 */
          struct vfs_ops tmp_ops;
         #endif
        #endif

        openlog("smbd_"VSCAN_MODULE_STR, LOG_PID, SYSLOG_FACILITY);

        #if SAMBA_VERSION_MAJOR==3
         /* Samba 3.0 alphaX */
         *vfs_version = SMB_VFS_INTERFACE_VERSION;
         vscan_syslog("samba-vscan (%s) loaded (Samba 3.x), (c) by Rainer Link, OpenAntiVirus.org", module_id);
        #else
         /* Samba 2.2.x */
         #if SAMBA_VERSION_RELEASE>=4
          /* Samba 2.2.4 */
          *vfs_version = SMB_VFS_INTERFACE_VERSION;
          vscan_syslog("samba-vscan (%s) loaded (Samba >=2.2.4), (c) by Rainer Link, OpenAntiVirus.org", module_id);
         #elif SAMBA_VERSION_RELEASE==2
          /* Samba 2.2.2 / Samba 2.2.3 !!! */
          *Version = SMB_VFS_INTERFACE_VERSION;
          vscan_syslog("samba-vscan (%s) loaded (Samba 2.2.2/2.2.3), (c) by Rainer Link, OpenAntiVirus.org", module_id);
         #else
          /* Samba 2.2.1 / Samba 2.2.0 */
          *vfs_version = SMB_VFS_INTERFACE_VERSION;
          vscan_syslog("samba-vscan (%s) loaded (Samba 2.2.0/2.2.1), (c) by Rainer Link, OpenAntiVirus.org",
               module_id);
         #endif
        #endif


        #if SAMBA_VERSION_MAJOR==3
         /* Samba 3.0 alphaX */
         DEBUG(3, ("Initialising default vfs hooks\n"));
         memcpy(&default_vfs_ops, def_vfs_ops, sizeof(struct vfs_ops));

         /* Remember vfs_handle for further allocation and referencing of
            private information in vfs_handle->data
         */
         vscan_handle = vfs_handle;
         return vscan_ops;
        #else
         /* Samba 2.2.x */
         #if SAMBA_VERSION_RELEASE>=4
          /* Samba 2.2.4 */

          *vfs_version = SMB_VFS_INTERFACE_VERSION;
          memcpy(&tmp_ops, def_vfs_ops, sizeof(struct vfs_ops));
          tmp_ops.connect = vscan_connect;
          tmp_ops.disconnect = vscan_disconnect;
          tmp_ops.open = vscan_open;
          tmp_ops.close = vscan_close;
          memcpy(&vscan_ops, &tmp_ops, sizeof(struct vfs_ops));
          return(&vscan_ops);

         #else
          /* Samba 2.2.3-2.2.0 */
          return(&vscan_ops);
         #endif
        #endif
}


#if SAMBA_VERSION_MAJOR==3
/* VFS finalization function */
void vfs_done(connection_struct *conn)
{
        DEBUG(3, ("Finalizing default vfs hooks\n"));
}
#endif

#endif /* #if (SMB_VFS_INTERFACE_VERSION >= 6) */



/*
 * init done
 */

void vfs_done(connection_struct * conn)
{
	DEBUG(3, ("Finalizing default vfs hooks\n"));
}
